import numpy as np

from ..abstract import Estimator
from . import Context


class TBATS(Estimator):
    """
    TBATS estimator used to fit and select best performing model.

    TBATS (Exponential smoothing state space model with Box-Cox
    transformation, ARMA errors, Trigonometric Trend and Seasonal components.)

    Model has been described in De Livera, Hyndman & Snyder (2011).

    All of the useful methods have been implemented in parent Estimator class.
    """

    def __init__(self,
                 use_box_cox=None, box_cox_bounds=(0, 1),
                 use_trend=None, use_damped_trend=None,
                 seasonal_periods=None, use_arma_errors=True,
                 show_warnings=True,
                 n_jobs=None, multiprocessing_start_method='spawn',
                 context=None):
        """ Class constructor

        Parameters
        ----------
        use_box_cox: bool or None, optional (default=None)
            If Box-Cox transformation of original series should be applied.
            When None both cases shall be considered and better is selected by AIC.
        box_cox_bounds: tuple, shape=(2,), optional (default=(0, 1))
            Minimal and maximal Box-Cox parameter values.
        use_trend: bool or None, optional (default=None)
            Indicates whether to include a trend or not.
            When None both cases shall be considered and better is selected by AIC.
        use_damped_trend: bool or None, optional (default=None)
            Indicates whether to include a damping parameter in the trend or not.
            Applies only when trend is used.
            When None both cases shall be considered and better is selected by AIC.
        seasonal_periods: iterable or array-like of floats, optional (default=None)
            Length of each of the periods (amount of observations in each period).
            TBATS accepts int and float values here.
            When None or empty array, non-seasonal model shall be fitted.
        use_arma_errors: bool, optional (default=True)
            When True BATS will try to improve the model by modelling residuals with ARMA.
            Best model will be selected by AIC.
            If False, ARMA residuals modeling will not be considered.
        show_warnings: bool, optional (default=True)
            If warnings should be shown or not.
            Also see Model.warnings variable that contains all model related warnings.
        n_jobs: int, optional (default=None)
            How many jobs to run in parallel when fitting BATS model.
            When not provided BATS shall try to utilize all available cpu cores.
        multiprocessing_start_method: str, optional (default='spawn')
            How threads should be started.
            See https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
        context: abstract.ContextInterface, optional (default=None)
            For advanced users only. Provide this to override default behaviors
        """
        if context is None:
            # the default TBATS context
            context = Context(show_warnings, n_jobs=n_jobs, multiprocessing_start_method=multiprocessing_start_method)
        super().__init__(context, use_box_cox=use_box_cox, box_cox_bounds=box_cox_bounds,
                         use_trend=use_trend, use_damped_trend=use_damped_trend,
                         seasonal_periods=seasonal_periods, use_arma_errors=use_arma_errors,
                         n_jobs=n_jobs)

    def _normalize_seasonal_periods(self, seasonal_periods):
        return self._normalize_seasonal_periods_to_type(seasonal_periods, dtype=float)

    def _do_fit(self, y):
        """Checks various model combinations to find best one by AIC"""
        components_grid = self._prepare_non_seasonal_components_grid()
        non_seasonal_model = self._choose_model_from_possible_component_settings(y, components_grid=components_grid)
        harmonics_choosing_strategy = self.context.create_harmonics_choosing_strategy()
        chosen_harmonics = harmonics_choosing_strategy.choose(y, self.create_most_complex_components())
        components_grid = self._prepare_components_grid(seasonal_harmonics=chosen_harmonics)
        seasonal_model = self._choose_model_from_possible_component_settings(y, components_grid=components_grid)

        if non_seasonal_model.aic < seasonal_model.aic:
            return non_seasonal_model

        return seasonal_model

    def create_most_complex_components(self):
        """Creates model components for the most complex model without ARMA residuals modelling"""
        components = dict(
            use_box_cox=self.use_box_cox is not False,
            use_trend=self.use_trend is not False,
            use_damped_trend=self.use_damped_trend is not False,
            seasonal_periods=self.seasonal_periods,
            use_arma_errors=False,
        )
        return self.context.create_components(**components)



from ..abstract import Estimator
from . import Context


class BATS(Estimator):
    """
    BATS estimator used to fit and select best performing model.

    BATS (Exponential smoothing state space model with Box-Cox
    transformation, ARMA errors, Trend and Seasonal components.)

    Model has been described in De Livera, Hyndman & Snyder (2011).

    All of the useful methods have been implemented in parent Estimator class.
    """

    def __init__(self,
                 use_box_cox=None, box_cox_bounds=(0, 1),
                 use_trend=None, use_damped_trend=None,
                 seasonal_periods=None, use_arma_errors=True,
                 show_warnings=True,
                 n_jobs=None, multiprocessing_start_method='spawn',
                 context=None):
        """ Class constructor

        Parameters
        ----------
        use_box_cox: bool or None, optional (default=None)
            If Box-Cox transformation of original series should be applied.
            When None both cases shall be considered and better is selected by AIC.
        box_cox_bounds: tuple, shape=(2,), optional (default=(0, 1))
            Minimal and maximal Box-Cox parameter values.
        use_trend: bool or None, optional (default=None)
            Indicates whether to include a trend or not.
            When None both cases shall be considered and better is selected by AIC.
        use_damped_trend: bool or None, optional (default=None)
            Indicates whether to include a damping parameter in the trend or not.
            Applies only when trend is used.
            When None both cases shall be considered and better is selected by AIC.
        seasonal_periods: iterable or array-like of int values, optional (default=None)
            Length of each of the periods (amount of observations in each period).
            BATS accepts only int values here.
            When None or empty array, non-seasonal model shall be fitted.
        use_arma_errors: bool, optional (default=True)
            When True BATS will try to improve the model by modelling residuals with ARMA.
            Best model will be selected by AIC.
            If False, ARMA residuals modeling will not be considered.
        show_warnings: bool, optional (default=True)
            If warnings should be shown or not.
            Also see Model.warnings variable that contains all model related warnings.
        n_jobs: int, optional (default=None)
            How many jobs to run in parallel when fitting BATS model.
            When not provided BATS shall try to utilize all available cpu cores.
        multiprocessing_start_method: str, optional (default='spawn')
            How threads should be started.
            See https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
        context: abstract.ContextInterface, optional (default=None)
            For advanced users only. Provide this to override default behaviors
        """
        if context is None:
            # the default BATS context
            context = Context(show_warnings, n_jobs=n_jobs, multiprocessing_start_method=multiprocessing_start_method)
        super().__init__(context, use_box_cox=use_box_cox, box_cox_bounds=box_cox_bounds,
                         use_trend=use_trend, use_damped_trend=use_damped_trend,
                         seasonal_periods=seasonal_periods, use_arma_errors=use_arma_errors,
                         n_jobs=n_jobs)

    def _normalize_seasonal_periods(self, seasonal_periods):
        """Makes sure periods are of int type"""
        return self._normalize_seasonal_periods_to_type(seasonal_periods, dtype=int)

    def _do_fit(self, y):
        """Checks all allowed parameter combinations in order to choose best model"""
        components_grid = self._prepare_components_grid()
        return self._choose_model_from_possible_component_settings(y, components_grid)



import math
import numpy as np

from ..abstract import Components, ComponentMatrix, SeedFinder as AbstractSeedFinder


class SeedFinder(AbstractSeedFinder):

    def __init__(self, components):
        """
        :param Components components:
        :return:
        """
        super().__init__(components)
        self.mask = None

    def to_matrix_for_linear_regression(self, w_tilda):
        w_tilda_obj = ComponentMatrix(w_tilda, self.components).without_arma()
        v, cuts = self.prepare_seasonal_params(w_tilda_obj)
        return w_tilda_obj.with_replaced_seasonal(v, cuts).as_matrix()

    def from_linear_regression_coefs_to_x0(self, linear_regression_coefs):
        cuts = self.mask_to_seasonal_cuts(self.get_mask())
        lr_coefs = ComponentMatrix(
            linear_regression_coefs,
            model_components=self.components.with_seasonal_periods(cuts),
            append_arma_coefs=True  # ARMA is not part of linear_regression so we need to re-add it
        )
        seasonal_part = self.transform_to_seasonal_vector(lr_coefs)
        x0 = lr_coefs.with_replaced_seasonal(seasonal_part, self.components.seasonal_periods).as_vector()
        return x0

    def prepare_seasonal_params(self, w_tilda_obj):
        w_tilda_seasons = w_tilda_obj.break_into_seasons()
        new_seasonal_params = []
        season = 0
        cuts = self.mask_to_seasonal_cuts(self.get_mask())
        for length in cuts:
            new_seasonal_params.append(w_tilda_seasons[season][:, 0:length])
            season += 1
        if len(new_seasonal_params) == 0:
            return np.empty((0, w_tilda_obj.matrix.shape[1])), cuts
        return np.block(new_seasonal_params), cuts

    def transform_to_seasonal_vector(self, lr_coefs):
        cuts = lr_coefs.components.seasonal_periods
        seasons = lr_coefs.break_into_seasons()
        new_seasonal_part = []
        for s in range(0, len(self.components.seasonal_periods)):
            season_length = self.components.seasonal_periods[s]
            season_coefs = seasons[s]
            cut_for_season = cuts[s]
            vec = np.asarray([0.0] * season_length)
            vec[:cut_for_season] = season_coefs
            vec = vec - np.mean(vec)  # remove mean so that it sums to 0
            new_seasonal_part.append(vec)
        if len(new_seasonal_part) == 0:
            return np.zeros((1, 0))
        return np.concatenate(new_seasonal_part)

    def get_mask(self):
        if self.mask is None:
            self.mask = self.prepare_mask(self.components.seasonal_periods)
        return self.mask

    @classmethod
    def prepare_mask(cls, seasonal_periods):
        periods = len(seasonal_periods)
        mask = [0] * periods
        for smaller in range(0, periods):  # for all seasonal periods
            for larger in range(smaller + 1, periods):  # for larger > smaller
                if (seasonal_periods[larger] % seasonal_periods[smaller]) == 0:
                    mask[smaller] = 1

        for larger in reversed(range(1, periods)):  # for larger in decreasing order
            for smaller in reversed(range(0, larger)):  # for smaller < larger in decreasing order
                if mask[smaller] == 1 or mask[larger] == 1:
                    continue
                hcf = math.gcd(seasonal_periods[smaller], seasonal_periods[larger])
                if hcf > 1:
                    mask[larger] = -hcf
        return mask

    def mask_to_seasonal_cuts(self, mask):
        seasonal_periods = self.components.seasonal_periods
        cuts = []
        # offset = 0
        for s in range(0, len(seasonal_periods)):
            period_length = seasonal_periods[s]
            if mask[s] < 0:  # there is a common divisor between seasons, cut out whole divisor
                length = period_length + mask[s]  # note that mask is negative
            elif mask[s] == 0:  # nothing in common, cut out last parameter
                length = period_length - 1
            else:  # when mask[s] == 1 cut whole season out as it is a sub-season of some longer season
                length = 0
            cuts.append(length)
            # offset += length

        return cuts



from ..abstract import ComponentMatrix, SeedFinder as AbstractSeedFinder
from . import Components


class SeedFinder(AbstractSeedFinder):

    def __init__(self, components):
        """
        :param Components components:
        :return:
        """
        super().__init__(components)
        self.mask = None

    def to_matrix_for_linear_regression(self, w_tilda):
        w_tilda_obj = ComponentMatrix(w_tilda, self.components).without_arma()
        return w_tilda_obj.as_matrix()

    def from_linear_regression_coefs_to_x0(self, linear_regression_coefs):
        lr_coefs = ComponentMatrix(
            linear_regression_coefs,
            model_components=self.components,
            append_arma_coefs=True  # ARMA is not part of linear_regression so we need to re-add it
        )
        return lr_coefs.as_vector()



from ..abstract import Case as AbstractCase


class Case(AbstractCase):

    def fit_initial_model(self, y):
        """Optimizes seasonal and non-seasonal models with no ARMA and returns the better one

        Parameters
        ----------
        y: array-like of floats
            Time series

        Returns
        -------
        Model
            The best model with no ARMA
        """
        model = self.fit_case(y, self.components.without_arma())
        if len(self.components.seasonal_periods) > 0:
            # Try non-seasonal model without ARMA
            model_candidate = self.fit_case(y, self.components.without_seasonal_periods().without_arma())
            if model_candidate.aic < model.aic:
                model = model_candidate
        return model



class ContextInterface(object):

    def get_exception_handler(self):
        # abstract method
        raise NotImplementedError()

    def create_constant_model(self, constant_value):
        # abstract method
        raise NotImplementedError()

    def create_model(self, params, validate_input=True):
        # abstract method
        raise NotImplementedError()

    def create_seed_finder(self, components):
        # abstract method
        raise NotImplementedError()

    def create_matrix_builder(self, params):
        # abstract method
        raise NotImplementedError()

    def create_default_starting_params(self, y, components):
        # abstract method
        raise NotImplementedError()

    def create_params_optimizer(self):
        # abstract method
        raise NotImplementedError()

    def create_case(self, components):
        # abstract method
        raise NotImplementedError()

    def create_case_from_dictionary(self, **components_dictionary):
        # abstract method
        raise NotImplementedError()

    def create_components(self, **components):
        # abstract method
        raise NotImplementedError()

    def create_harmonics_choosing_strategy(self):
        # abstract method
        raise NotImplementedError()

    def multiprocessing(self):
        # abstract method
        raise NotImplementedError()


import numpy as np
import fractions
import multiprocessing


class HarmonicsChoosingStrategy(object):

    def __init__(self, context, checking_range=None):
        self.checking_range = checking_range
        if checking_range is None:
            self.checking_range = multiprocessing.cpu_count()
        self.context = context

    def choose(self, y, components):
        if len(components.seasonal_periods) == 0:
            return np.asarray([])

        components = components.with_harmonics_as_ones()
        best_model = self.context.create_case(components).fit(y)

        max_harmonics = self.calculate_max(components.seasonal_periods)

        for season_index in range(0, len(components.seasonal_periods)):
            best_model = self.choose_for_season(
                season_index=season_index,
                max_harmonic=max_harmonics[season_index],
                best_model_so_far=best_model
            )
        return best_model.params.components.seasonal_harmonics

    def initial_harmonics_to_check(self, max_harmonic):
        first_harmonics_to_check = 6
        harmonic_to_check = np.min([max_harmonic, first_harmonics_to_check])
        parallel = np.max([self.checking_range, 3])
        range_max = harmonic_to_check + int(np.ceil(parallel / 2))
        range_max = np.max([range_max, 2 + parallel])
        range_max = np.min([range_max, max_harmonic + 1])
        return range(
            np.max([range_max - parallel, 2]),
            range_max
        )

    def next_harmonics_to_check(self, max_harmonic, previously_checked=None, chosen_harmonic=None):
        if previously_checked is None or len(previously_checked) == 0:
            return self.initial_harmonics_to_check(max_harmonic)
        if chosen_harmonic < max_harmonic and chosen_harmonic == np.max(previously_checked):
            # we have chosen the most complex model, we need to check even more complex ones
            return range(
                chosen_harmonic + 1,
                np.min([chosen_harmonic + self.checking_range, max_harmonic]) + 1
            )

        if chosen_harmonic > 2 and chosen_harmonic == np.min(previously_checked):
            # we have chosen the least complex model, we need to check even simpler models
            return range(
                np.max([2, chosen_harmonic - self.checking_range]),
                chosen_harmonic
            )
        if chosen_harmonic == 1 and np.min(previously_checked) > 2 and max_harmonic > 1:
            # we are still on the initial model, check simple models
            return range(
                np.max([2, np.min(previously_checked) - self.checking_range]),
                np.min(previously_checked)
            )
        return []

    def choose_for_season(self, season_index, max_harmonic, best_model_so_far):
        # assertion: best_model_so_far uses harmonics=1 for the seasonality being analysed
        if max_harmonic == 1:  # nothing to search
            return best_model_so_far

        best_model = best_model_so_far
        best_aic = np.inf
        previously_checked_harmonics = np.asarray([])
        harmonics_range = self.next_harmonics_to_check(max_harmonic)

        while len(harmonics_range) > 0:
            self._season_index = season_index
            self._y = best_model_so_far.y
            self._components = best_model_so_far.params.components
            with self.context.multiprocessing().Pool(processes=self.checking_range) as pool:
                models = pool.map(self._fit_model, harmonics_range)
                pool.close()
            for model in models:
                if model.aic < best_aic:
                    best_model = model
                    best_aic = model.aic
            chosen_harmonics = best_model.params.components.seasonal_harmonics[season_index]
            previously_checked_harmonics = np.concatenate([previously_checked_harmonics, harmonics_range])
            harmonics_range = self.next_harmonics_to_check(
                max_harmonic=max_harmonic,
                previously_checked=previously_checked_harmonics,
                chosen_harmonic=chosen_harmonics,
            )

        # cleanup
        self._y = None
        self._season_index = None
        self._components = None

        if best_aic > best_model_so_far.aic:
            return best_model_so_far
        return best_model

    def _fit_model(self, harmonic_to_check):
        components = self._components.with_harmonic_for_season(
            season_index=self._season_index, new_harmonic=harmonic_to_check
        )
        return self.context.create_case(components=components).fit(self._y)

    def choose_for_season_serial(self, season_index, max_harmonic, best_model_so_far):
        # assertion: best_model_so_far uses harmonics=1 for the seasonality being analysed

        if max_harmonic == 1:  # nothing to search
            return best_model_so_far

        first_harmonics_to_check = 6
        search_direction = -1  # -1 we will be reducing harmonics, 1 we shall be increasing it

        harmonic_to_check = np.min([max_harmonic, first_harmonics_to_check])

        best_model = best_model_so_far
        best_model_aic = np.inf
        if max_harmonic > first_harmonics_to_check:
            level_model = self.fit_model_like_previous_with_harmonic(
                best_model_so_far, season_index, first_harmonics_to_check
            )
            up_model = self.fit_model_like_previous_with_harmonic(
                best_model_so_far, season_index, first_harmonics_to_check + 1
            )
            down_model = self.fit_model_like_previous_with_harmonic(
                best_model_so_far, season_index, first_harmonics_to_check - 1
            )
            # by default we go into less complex models
            search_direction = -1
            harmonic_to_check = first_harmonics_to_check - 1
            best_model = level_model
            best_model_aic = level_model.aic

            if down_model.aic < best_model_aic:
                best_model = down_model
                best_model_aic = down_model.aic
                harmonic_to_check = first_harmonics_to_check - 2

            if up_model.aic < best_model_aic:
                # in such a case we shall go into more complex models
                search_direction = 1
                best_model = up_model
                best_model_aic = up_model.aic
                harmonic_to_check = first_harmonics_to_check + 2

        while harmonic_to_check > 1 and harmonic_to_check <= max_harmonic:
            candidate_model = self.fit_model_like_previous_with_harmonic(
                best_model_so_far, season_index, harmonic_to_check
            )
            if candidate_model.aic > best_model_aic:
                # AIC stopped getting better
                break
            best_model = candidate_model
            best_model_aic = candidate_model.aic
            harmonic_to_check += search_direction

        if best_model_aic < best_model_so_far.aic:
            return best_model

        return best_model_so_far

    def fit_model_like_previous_with_harmonic(self, previous_model, season_index, harmonic_to_check):
        components = previous_model.params.components.with_harmonic_for_season(
            season_index=season_index, new_harmonic=harmonic_to_check
        )
        return self.context.create_case(components=components).fit(previous_model.y)

    @classmethod
    def calculate_max(cls, seasonal_periods, dependency_reduction_function=None):
        if dependency_reduction_function is None:
            # should this be used or _better version?
            dependency_reduction_function = cls.max_harmonic_dependency_reduction
        max_harmonics = [1] * len(seasonal_periods)
        for period_index in range(0, len(seasonal_periods)):
            period_length = seasonal_periods[period_index]

            # for 2,3,4 -> 1; for 5,6 -> 2; for 7,8 -> 3; ...
            max_harmonic = np.max([1, int((period_length - 1) / 2)])
            max_harmonic = dependency_reduction_function(
                max_harmonic_proposal=max_harmonic,
                period_length=period_length,
                seasonal_periods=seasonal_periods
            )
            max_harmonics[period_index] = max_harmonic

        return max_harmonics

    @classmethod
    def max_harmonic_dependency_reduction(cls, max_harmonic_proposal, period_length, seasonal_periods):
        if period_length % 1 != 0:
            # no dependendencies if period length is float
            return max_harmonic_proposal

        shorter_seasonal_periods = seasonal_periods[np.where(seasonal_periods < period_length)]
        max_harmonic = max_harmonic_proposal
        # reduce max harmonic if there is a shorter season
        # that would otherwise repeat the same seasonal trigonometric components
        harmonic = 2
        while harmonic <= max_harmonic:
            if period_length % harmonic == 0:
                amount_of_times_fits_in_period = period_length / harmonic  # this will be integer
                if np.any(shorter_seasonal_periods % amount_of_times_fits_in_period == 0):
                    max_harmonic = harmonic - 1
            harmonic += 1

        return max_harmonic

    @classmethod
    def max_harmonic_dependency_reduction_better(cls, max_harmonic_proposal, period_length, seasonal_periods):
        shorter_seasonal_periods = seasonal_periods[np.where(seasonal_periods < period_length)]
        for shorter_period_length in shorter_seasonal_periods:
            # we need to do type casting here as np.float and np.int64 do not behave well in fractions package
            real_ratio = float(shorter_period_length / period_length)
            # approximate a potentially real number with a rational number
            fraction = fractions.Fraction(real_ratio).limit_denominator(int(period_length) + 1)
            rational_ratio_approximation = fraction.numerator / fraction.denominator

            if not np.isclose(real_ratio, rational_ratio_approximation, atol=1e-5):
                # approximation is bad, no dependency between periods
                continue

            # fraction denominator points to the first harmonic in longer period
            # that shall repeat harmonic in shorter period (being pointed in fraction nominator)
            max_harmonic_proposal = np.min([fraction.denominator - 1, max_harmonic_proposal])

        return max_harmonic_proposal
